use clap::{Parser,Subcommand};
use crate::{schema,cert};

#[derive(Debug, Parser)]
#[clap(author, version, about, long_about = None)]
struct RootArgs {
    #[clap(default_value_t = String::from("vogt.db"))]
    db: String,

    #[clap(short, long, default_value_t = String::from("primary"))]
    realm: String,

    #[clap(subcommand)]
    command: Command,
}

#[derive(Debug,Subcommand)]
enum Command {
    Init,
    Cert(CertArgs),
    Group(GroupArgs),
    Server(ServerArgs),
    Token(TokenArgs),
    User(UserArgs),
}

impl RootArgs {
    async fn run(&self) {
        if let Command::Init = self.command {
            return self.init().await;
        }

        let storage = microrm::DB::new(schema::schema(), &self.db, microrm::CreateMode::MustExist);

        if let Err(e) = storage {
            println!("Error occured while loading database: {}", e);
            return
        }
        let storage = storage.unwrap();

        match &self.command {
            Command::Init => unreachable!(),
            Command::Cert(v) => v.run(&self, &storage).await,
            Command::Group(v) => v.run(&self, &storage).await,
            Command::Server(v) => v.run(&self, &storage).await,
            Command::Token(v) => v.run(&self, &storage).await,
            Command::User(v) => v.run(&self, &storage).await,
        }
    }

    async fn init(&self) {
        // first check to see if the database is already vaguely set up
        let maybedb = microrm::DB::new(schema::schema(), &self.db, microrm::CreateMode::MustExist);

        if maybedb.is_ok() {
            println!("Database already initialized, not overwriting!");
            return
        }

        println!("Initializing!");

        let db = microrm::DB::new(schema::schema(), &self.db, microrm::CreateMode::AllowNewDatabase).expect("Unable to initialize database!");

        // create primary realm
        let qi = db.query_interface();
        qi.add(&schema::Realm { shortname: "primary".to_string() }).expect("couldn't add primary realm");
    }
}

#[derive(Debug,Subcommand)]
enum CertCommand {
    Inspect,
    Generate,
}

#[derive(Debug, Parser)]
struct CertArgs {
    #[clap(subcommand)]
    command: CertCommand
}

impl CertArgs {
    async fn run(&self, root: &RootArgs, si: &microrm::DB) {
        match &self.command {
            CertCommand::Inspect => {
                cert::inspect(si, &root.realm);
            }
            CertCommand::Generate => {
                cert::generate(si, &root.realm);
            }
        }
    }
}

#[derive(Debug,Subcommand)]
enum GroupCommand {
}

#[derive(Debug, Parser)]
struct GroupArgs {
    #[clap(subcommand)]
    command: GroupCommand
}

impl GroupArgs {
    async fn run(&self, root: &RootArgs, si: &microrm::DB) {
    }
}


#[derive(Debug, Parser)]
struct ServerArgs {
    port: Option<u16>,
}

impl ServerArgs {
    async fn run(&self, root: &RootArgs, db: &microrm::DB) {
        crate::server::launch(db, self.port.unwrap_or(2000)).await
    }
}

#[derive(Debug, Subcommand)]
enum TokenCommand {
    GenerateAuth { username: String, scopes: String },
    GenerateRefresh { username: String, scopes: String },
    Inspect { token: String },
}

#[derive(Debug, Parser)]
struct TokenArgs {
    #[clap(subcommand)]
    command: TokenCommand,
}

impl TokenArgs {
    async fn run(&self, root: &RootArgs, _db: &microrm::DB) {
        
    }
}

#[derive(Debug, Subcommand)]
enum UserCommand {
    Create { username: String },
    List { filter: Option<String> },
}

#[derive(Debug, Parser)]
struct UserArgs {
    #[clap(subcommand)]
    command: UserCommand,
}

impl UserArgs {
    async fn run(&self, root: &RootArgs, _db: &microrm::DB) {
        
    }
}

pub fn invoked() {
    let args = RootArgs::parse();

    smol::block_on(args.run());
    // async_std::task::block_on(args.run());
}
